#region
/* Adoor.NET - Object/Relational mapping framework leveraging ADO.NET`s powerful disconnected mode.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *  
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *  
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */ 
#endregion

using System;
using System.Collections.Generic;
using Adoor.Expression.AST;

namespace Adoor.Expression
{
	/// <summary>
	/// Description rsume de 
	/// </summary>
	public class ExpressionParser
	{
		public ExpressionNode ParseCondition(string condition)
		{
			Scanner scanner = new Scanner(condition);
			scanner.Scan();
			ExpressionNode result = ParseSearchCondition(scanner);
			if (scanner.Token != ScannerToken.End)
				throw new Exception("syntax error");
			return result;
		}

		public ExpressionNode ParseExpression(string condition)
		{
			Scanner scanner = new Scanner(condition);
			scanner.Scan();
			ExpressionNode result = ParseRowValueConstructor(scanner);
			if (scanner.Token != ScannerToken.End)
				throw new Exception("syntax error");
			return result;
		}

		// <search condition> ::=
		//		<boolean term>
		//		| <search condition> OR <boolean term>
		public ExpressionNode ParseSearchCondition(Scanner scanner)
		{
			ExpressionNode result = ParseBooleanTerm(scanner);
			while (scanner.Token == ScannerToken.Or)
			{
				scanner.Scan();
				result = new ExpressionBinary(ScannerToken.Or, result, ParseBooleanTerm(scanner));
			}
			return result;
		}

		// <boolean term> ::=
		//		<boolean factor>
		//		| <boolean term> AND <boolean factor>
		private ExpressionNode ParseBooleanTerm(Scanner scanner)
		{
			ExpressionNode result = ParseBooleanFactor(scanner);
			while (scanner.Token == ScannerToken.And)
			{
				scanner.Scan();
				result = new ExpressionBinary(ScannerToken.And, result, ParseBooleanFactor(scanner));
			}
			return result;
		}

		// <boolean factor> ::=
		//		[ NOT ] <boolean test>
		private ExpressionNode ParseBooleanFactor(Scanner scanner)
		{
			
			if (scanner.Token != ScannerToken.Not)
				return ParseBooleanTest(scanner);
			scanner.Scan();
			return new ExpressionUnary(ScannerToken.Not, ParseBooleanTest(scanner));
		}

		// <boolean test> ::=
		//		<boolean primary> [ IS [ NOT ] <truth value> ]
		private ExpressionNode ParseBooleanTest(Scanner scanner)
		{

			ExpressionNode result = ParseBooleanPrimary(scanner);
			if (scanner.Token == ScannerToken.Is)
			{
				scanner.Scan();
				bool isNot = false;
				if (scanner.Token == ScannerToken.Not)
				{
					scanner.Scan();
					isNot = true;
				}
				if (scanner.Token != ScannerToken.True && scanner.Token != ScannerToken.False)
					throw new Exception("'TRUE' or 'FALSE' expected");  // 'UNKNOWN' ?
				ScannerToken t = scanner.Token;
				scanner.Scan();
				result = new ExpressionIs(result, isNot, t);
			}

			return result;
		}

		// <boolean primary> ::=
		//		<predicate>
		//		| <left paren> <search condition> <right paren>
		private ExpressionNode ParseBooleanPrimary(Scanner scanner)
		{
			ExpressionNode result = null;
			if (scanner.Token != ScannerToken.LeftParen)
			{
				return ParsePredicate(scanner);
			}
			scanner.Scan();
			
			result = ParseSearchCondition(scanner);

			if (scanner.Token != ScannerToken.RightParen)
			{
				throw new Exception("Parenthesis error");
			}
			scanner.Scan();
			result = new ExpressionParen(result);			

			while(scanner.Token == ScannerToken.Asterisk || scanner.Token == ScannerToken.Solidus)
			{
				ScannerToken t = scanner.Token;
				scanner.Scan();
				result = new ExpressionBinary(t, result, ParseFactor(scanner));
			}

			while(scanner.Token == ScannerToken.PlusSign 
				|| scanner.Token == ScannerToken.MinusSign
				|| scanner.Token == ScannerToken.BitWiseAnd
				|| scanner.Token == ScannerToken.BitWiseOr
				|| scanner.Token == ScannerToken.Modulo)
			{
				ScannerToken t = scanner.Token;
				scanner.Scan();
				result = new ExpressionBinary(t, result, ParseValueExpression(scanner));				
			}

			return result;
		}

		// <predicate> ::=
		//		<comparison predicate>
		//		| <like predicate>
		//		| <null predicate>
		// <comparison predicate> ::=
		//		<row value constructor> <comp op>
		//		<row value constructor>
		//	<like predicate> ::=
		//		<match value> [ NOT ] LIKE <pattern>  -- <match value> -> <row value constructor>, <pattern> -> <row value constructor>
		//		[ ESCAPE <escape character> ]
		// <null predicate> ::= <row value constructor>
		//		IS [ NOT ] NULL
		private ExpressionNode ParsePredicate(Scanner scanner)
		{
			if( scanner.Token ==  ScannerToken.Exists)
			{
				scanner.Scan();
				if(scanner.Token != ScannerToken.LeftParen)				
					throw new Exception("'(' expected");

				scanner.Scan();
				
				if(scanner.Token != ScannerToken.Select)
					throw new Exception("select expected");
					
				object subQuery = OnParseSubQuery(scanner, true);
					
				if(scanner.Token != ScannerToken.RightParen)
					throw new Exception("')' expected");
				scanner.Scan();
				return new ExpressionExistsSubquery(subQuery);				
			}
			else
			{
				ExpressionNode result = ParseRowValueConstructor(scanner);				

				if (scanner.Token == ScannerToken.RightParen 
					|| scanner.Token == ScannerToken.Comma)
					return result;
				else if (IsCompOp(scanner.Token)) 
				{
					ScannerToken compOp = scanner.Token;
					scanner.Scan();
					ExpressionNode rowValueConstructor = ParseRowValueConstructor(scanner);

					return new ExpressionBinary(compOp, result, rowValueConstructor);
				}
				else if (scanner.Token == ScannerToken.Not) 
				{
					scanner.Scan();
					if (scanner.Token != ScannerToken.Like)
						throw new Exception("'like' expected");
					scanner.Scan();
					return new ExpressionLike(result, true, ParseRowValueConstructor(scanner));
				}
				else if (scanner.Token == ScannerToken.Like)
				{
					scanner.Scan();
					return new ExpressionLike(result, false, ParseRowValueConstructor(scanner));
				}
				else if (scanner.Token == ScannerToken.Is)
				{
					ScannerToken oldtoken = scanner.Token;
					int oldpos = scanner.Pos;
					bool isNotNull = false;
					scanner.Scan();
					if (scanner.Token == ScannerToken.Not)
					{
						scanner.Scan();
						isNotNull = true;
					}
					if (scanner.Token != ScannerToken.Null) 
					{
						// backup to oldpos: it can be IS [NOT] (TRUE|FALSE)
						scanner.Pos = oldpos;
						scanner.Token = oldtoken;
						return result;
					}
					scanner.Scan();
					return new ExpressionIs(result, isNotNull, ScannerToken.Null);
				}
				else if (scanner.Token == ScannerToken.In)
				{				
					scanner.Scan();
					if(scanner.Token != ScannerToken.LeftParen)				
						throw new Exception("'(' expected");

                    List<ExpressionNode> alExpr = new List<ExpressionNode>();
					do
					{
						scanner.Scan();
						if(scanner.Token == ScannerToken.Select)
						{						
							object subQuery = OnParseSubQuery(scanner, false);
							if(scanner.Token != ScannerToken.RightParen)
								throw new Exception("')' expected");
							scanner.Scan();
							return new ExpressionInSubquery(result, subQuery);						
						}

						ExpressionNode term = ParseTerm(scanner);
						alExpr.Add(term);
					} while(scanner.Token == ScannerToken.Comma);

					if(alExpr.Count == 0)
						throw new Exception(" in cannot have empty list of values");
				
					if(scanner.Token != ScannerToken.RightParen)
						throw new Exception("')' expected");
					scanner.Scan();

					return new ExpressionIn(result, alExpr.ToArray());
				}
				else 
					throw new Exception("'not', 'like' or 'is' or 'in' expected");
			}
		}

		public delegate void ParseSubQueryEventHandler(object sender, ParseSubQueryEventArgs e);

		public class ParseSubQueryEventArgs : EventArgs
		{
			public ParseSubQueryEventArgs(Scanner scanner, bool forwardCorrelations)
			{
				this.scanner = scanner;
				this.forwardCorrelations = forwardCorrelations;
			}

			private Scanner scanner;
			public Scanner Scanner
			{
				get{ return scanner;}
			}

			private object subQuery;
			public object SubQuery
			{
				get{ return subQuery; }
				set{ subQuery = value; }
			}

			private object sourceQuery;
			public object SourceQuery
			{
				get{ return sourceQuery; }
				set{ sourceQuery = value; }
			}

			private bool forwardCorrelations;
			public bool ForwardCorrelations
			{
				get{ return forwardCorrelations; }
				set{ forwardCorrelations = value; }
			}
		}

		protected object OnParseSubQuery(Scanner scanner, bool forwardCorrelations)
		{
			ParseSubQueryEventArgs e = new ParseSubQueryEventArgs(scanner, forwardCorrelations);
			if(ParseSubQuery != null)
				ParseSubQuery(this, e);
			if(e.SubQuery == null)
				throw new NullReferenceException("SubQuery cannot be null");
			return e.SubQuery;
		}

		public event ParseSubQueryEventHandler ParseSubQuery;

		// <row value constructor element> ::=
		//		<value expression>
		//		| <null specification>
		public ExpressionNode ParseRowValueConstructor(Scanner scanner)
		{
			if (scanner.Token == ScannerToken.Null)
			{
				scanner.Scan();
				return new ExpressionPrimary(ScannerToken.Null, "null");
			}
			ExpressionNode result = ParseValueExpression(scanner);
			if (scanner.Token == ScannerToken.LeftBrace)
			{
				if(!(result is ExpressionName))
					throw new Exception("syntax error");
				result = (new ExpressionParser()).ParseConditonalNode(result as ExpressionName, scanner);
			}
			return result;
		}

		public ExpressionNode ParseConditonalNode(ExpressionName qualifier, Scanner scanner)
		{
			ExpressionNode defaultNode = null;

            Dictionary<string, ExpressionNode> nodes =
                new Dictionary<string, ExpressionNode>(StringComparer.CurrentCultureIgnoreCase);
			do
			{
				int start = scanner.Pos;
				scanner.Scan();
				if(scanner.Token == ScannerToken.Name)
				{
					string correlation = scanner.GetString();
					scanner.Scan();
					if(scanner.Token == ScannerToken.Colon)
					{						
						scanner.Scan();
						ExpressionNode childNode = ParseRowValueConstructor(scanner);
						nodes.Add(correlation, childNode);
					}
					else if(scanner.Token == ScannerToken.Period)
					{
						scanner.Pos = start;
						scanner.Scan();
						ExpressionNode childNode = ParseRowValueConstructor(scanner);
						nodes.Add(correlation, childNode);				
					}
					else
					{
						throw new Exception("syntax error");
					}
				}
				else if(defaultNode == null)
				{
					defaultNode = ParseRowValueConstructor(scanner);
				}
				else
				{
					throw new Exception("syntax error");
				}
			}while(scanner.Token == ScannerToken.Comma);

			if(scanner.Token != ScannerToken.RightBrace)
				throw new Exception("'}' expected ");

			scanner.Scan();

			if(defaultNode == null)
				defaultNode = new ExpressionPrimary(ScannerToken.Null, "NULL");

			return new ExpressionCorrelationFilteredNode(qualifier, nodes, defaultNode);
		}

		// <value expression> ::=
		//		<numeric value expression>
		// <numeric value expression> ::=
		//		<term>
		//		| <numeric value expression> <plus sign> <term>
		//		| <numeric value expression> <minus sign> <term>
		private ExpressionNode ParseValueExpression(Scanner scanner)
		{
			ExpressionNode result = null;
			if(scanner.Token == ScannerToken.LeftParen)
			{
				scanner.Scan();
				result = ParseValueExpression(scanner);

				if (scanner.Token != ScannerToken.RightParen)
					throw new Exception(" ')' error");
				scanner.Scan();

				result = new ExpressionParen(result);
				while(scanner.Token == ScannerToken.Asterisk || scanner.Token == ScannerToken.Solidus)
				{
					ScannerToken t = scanner.Token;
					scanner.Scan();
					result = new ExpressionBinary(t, result, ParseFactor(scanner));
				}
			}
			else
				result = ParseTerm(scanner);
			while(scanner.Token == ScannerToken.PlusSign 
				|| scanner.Token == ScannerToken.MinusSign
				|| scanner.Token == ScannerToken.BitWiseAnd
				|| scanner.Token == ScannerToken.BitWiseOr
				|| scanner.Token == ScannerToken.Modulo)
			{
				ScannerToken t = scanner.Token;
				scanner.Scan();
				result = new ExpressionBinary(t, result, ParseValueExpression(scanner));				
			}
			return result;
		}

		// <term> ::=
		//		<factor>
		//		| <term> <asterisk> <factor>
		//		| <term> <solidus> <factor>
		private ExpressionNode ParseTerm(Scanner scanner)
		{
			ExpressionNode result = ParseFactor(scanner);
			while(scanner.Token == ScannerToken.Asterisk || scanner.Token == ScannerToken.Solidus)
			{
				ScannerToken t = scanner.Token;
				scanner.Scan();
				result = new ExpressionBinary(t, result, ParseFactor(scanner));
			}
			return result;
		}

		//<factor> ::=
		//		[ <sign> ] <numeric primary>
		private ExpressionNode ParseFactor(Scanner scanner)
		{
			if (scanner.Token == ScannerToken.PlusSign 
				|| scanner.Token == ScannerToken.MinusSign
				|| scanner.Token == ScannerToken.BitWiseNot)
			{
				ScannerToken t = scanner.Token;
				scanner.Scan();
				return new ExpressionUnary(t, ParseNumericPrimary(scanner));
			}
			return ParseNumericPrimary(scanner);
		}

		public ExpressionNode ParseFunction(string name, Scanner scanner)
		{
			if (scanner.Token == ScannerToken.LeftParen)
			{
				scanner.Scan();
                List<ExpressionNode> parameters = new List<ExpressionNode>();
				if (scanner.Token != ScannerToken.RightParen) 
				{
					do
					{								
						if(scanner.Token == ScannerToken.Comma)
							scanner.Scan();

						ExpressionNode expr = null;
						expr = ParseBooleanTerm(scanner);
						parameters.Add(expr);
								
					}while (scanner.Token == ScannerToken.Comma);
				}
				if (scanner.Token != ScannerToken.RightParen)
					throw new Exception("')' expected");

				scanner.Scan();
				return new ExpressionFunction(name, parameters.ToArray());
			}
			throw new Exception();
		}

		// <numeric primary> ::=
		//		<value expression primary>
		// <value expression primary> ::=
		//		<unsigned value specification>
		//		| <column reference>
		//		| <set function specification>
		//		| <left paren> <value expression> <right paren>	-- OK
		// <unsigned value specification> ::=
		//		<unsigned literal>
		//		| <general value specification>
		// <unsigned literal> ::=
		//		<unsigned numeric literal>
		//		| <general literal>
		// <general value specification> ::=
		//		<parameter specification>
		//		| <dynamic parameter specification>
		// <parameter specification> ::=
		//		<parameter name>
		// <parameter name> ::= <colon> <identifier>
		// <column reference> ::= [ <qualifier> <period> ] <column name>
		// <general literal> ::=
		//		<character string literal>
		//		| <national character string literal>
		//		| <bit string literal>
		//		| <hex string literal>
		//		| <datetime literal>
		//		| <interval literal>
		public ExpressionNode ParseNumericPrimary(Scanner scanner)
		{
			ExpressionNode result;
			switch (scanner.Token)
			{
				case ScannerToken.LeftParen:
					return ParseValueExpression(scanner);					
				case ScannerToken.Distinct:
				case ScannerToken.Name:
					string name = scanner.GetString();
					// <column reference>
					//	| <set function specification>
					scanner.Scan();
					if(name.StartsWith("@"))
						return new ExpressionPrimary(ScannerToken.Params, name);

					if (scanner.Token == ScannerToken.Period)
					{
						scanner.Scan();
						if (scanner.Token == ScannerToken.Asterisk)
						{
							scanner.Scan();
							return new ExpressionName(name, "*");
						}
						if (scanner.Token != ScannerToken.Name)
							throw new Exception("name expected");
						string value = scanner.GetString();
						scanner.Scan();
						if(scanner.Token == ScannerToken.LeftParen)
							return ParseFunction(name +"."+ value, scanner);
						else
							return new ExpressionName(name, value);
					}
					if (scanner.Token == ScannerToken.LeftParen)
					{
						return ParseFunction(name, scanner);
					}
					return new ExpressionName("", name);
				case ScannerToken.Date:
				case ScannerToken.Integer:
				case ScannerToken.Single:
				case ScannerToken.Double:
				case ScannerToken.String:
				case ScannerToken.QuestionMark:
					result = new ExpressionPrimary(scanner.Token, scanner.GetString());
					scanner.Scan();
					return result;
				case ScannerToken.LeftBrace:
					scanner.Scan();
					int oldstart = scanner.Start;
					while (scanner.Token != ScannerToken.RightBrace)
					{
						if (scanner.Token == ScannerToken.End)
							throw new Exception("'}' expected");
						scanner.Scan();
					}
					string content = scanner.GetString(oldstart, scanner.Pos - oldstart - 1);
					scanner.Scan();
					return new ExpressionEscape(content);
				default:
					throw new Exception("syntax error");
			}
		}

		protected bool IsCompOp(ScannerToken t)
		{
			switch (t)
			{
				case ScannerToken.Eq:
				case ScannerToken.Ne:
				case ScannerToken.Gt:
				case ScannerToken.Ge:
				case ScannerToken.Lt:
				case ScannerToken.Le:
					return true;
			}
			return false;
		}

		private object tag;

		public object Tag
		{
			get{ return tag; }
			set{ tag = value; }
		}
	}
}
