#region
/* Adoor.NET - Object/Relational mapping framework leveraging ADO.NET`s powerful disconnected mode.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *  
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *  
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#endregion

#define CSHARP_COMPILER
#define DEBUG_COMPILER

using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Security.Permissions;
using System.Xml;
using Adoor.Object.Domain;
using Adoor.Object.Entity;
using Microsoft.CSharp;
using Microsoft.VisualBasic;

namespace Adoor.Object.Domain.DomainBuild.CodeDom
{
    /// <summary>
    /// Internal object used to build <see cref="ObjectDomain"/> objects.
    /// </summary>
    public class ObjectDomainBuilderOld
    {
        //protected static log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        /// <summary>
        /// Creates a new <see cref="ObjectDomain"/>.
        /// </summary>
        /// <param name="assemblyName">The assembly containing the entity classes for this domain.</param>
        /// <param name="mapFileName">The name of the xml file. If the xml file is embeded as a resource inside an assembly, the name of the assembly followed by the "#" sign should prefix the resource name.</param>
        /// <param name="extendedProperties">A dictionary of extended properties.</param>
        public static void Build(ObjectDomain domain, string assemblyName, string mapFileName)
        {
            Assembly assembly = null;
            if (assemblyName != null && assemblyName != string.Empty)
                assembly = Assembly.Load(assemblyName);

            XmlDocument xmlMap = new XmlDocument();

            int pos = mapFileName.IndexOf("#");
            Stream stream;
            if (pos == -1)
                stream = new FileStream(mapFileName, FileMode.Open, FileAccess.Read);
            else
            {
                string mapAssemblyName = mapFileName.Substring(0, pos);
                string mapResourceName = mapFileName.Substring(pos + 1);
                Assembly mapAssembly = Assembly.Load(mapAssemblyName);
                stream = mapAssembly.GetManifestResourceStream(mapResourceName);
            }
            if (stream == null)
                throw new Exception("Map file " + mapFileName + " not found");
            using (stream)
            {
                xmlMap.Load(new XmlTextReader(stream));
            }
            stream.Close();

            XmlNamespaceManager nsmgr = new XmlNamespaceManager(new NameTable());
            nsmgr.AddNamespace("typing", "urn:xmd-typing");
            nsmgr.AddNamespace("mapping", "urn:xmd-mapping");

            try
            {
                Assembly implementationAssembly = Adoor.Object.Domain.DomainBuild.CodeGeneration.CodeGeneratorCodeDom.GetImplementationAssembly(assembly, xmlMap, nsmgr, assembly == null);

                ReadMap(domain, implementationAssembly, xmlMap, nsmgr, assembly == null);

                foreach (EntityData entityData in domain.EntityInfos)
                {
                    foreach (ReferenceData referenceData in entityData.ReferenceInfos.Values)
                    {
                        IInverseReference inverseReference = (IInverseReference)referenceData.GetService(typeof(IInverseReference));
                        if (inverseReference != null)
                            inverseReference.InitInverseReference();
                    }
                }

                foreach (EntityData entityData in domain.EntityInfos)
                {
                    IService service = (IService)entityData.GetService(typeof(IService));
                    if (service != null)
                        service.Initialize();

                    foreach (PropertyData propertyData in entityData.PropertyInfos.Values)
                        initProperty(propertyData);

                    foreach (ReferenceData referenceData in entityData.ReferenceInfos.Values)
                        initReference(referenceData);
                }
            }
            catch (Exception e)
            {
                throw new Exception("Error reading map file " + mapFileName + ": " + e.Message, e);
            }
        }

        private static void ReadMap(ObjectDomain domain, Assembly implementationAssembly, XmlNode xmlMap, XmlNamespaceManager nsmgr, bool generate)
        {
            //XmlNodeList nodeList = xmlMap.SelectNodes("//*/mapping:entity/..", nsmgr);
            // We need to sort nodes for inheritance issues

            List<XmlNode> alOrderedNodes = new List<XmlNode>();

            XmlNodeList nodeList = xmlMap.SelectNodes("//*[count(typing:entity/@parent) = 0]/mapping:entity/..", nsmgr);
            foreach (XmlNode o in nodeList)
                alOrderedNodes.Add(o);

            Dictionary<string, XmlElement> ht = new Dictionary<string, XmlElement>();
            nodeList = xmlMap.SelectNodes("//*[count(typing:entity/@parent) = 1]", nsmgr);
            foreach (XmlElement elem in nodeList)
            {
                ht[elem.SelectSingleNode("typing:entity/@type", nsmgr).Value] = elem;
            }

            IEnumerator<string> it = null;
            string type = null;
            //TODO: Possible infinite loop ? 
            while (ht.Count > 0)
            {
                if (type == null)
                {
                    it = ht.Keys.GetEnumerator();
                    if (it.MoveNext())
                    {
                        type = (string)it.Current;
                    }
                }
                else
                {
                    XmlElement elem = ht[type];
                    string parent = elem.SelectSingleNode("typing:entity/@parent", nsmgr).Value;
                    if (ht.ContainsKey(parent))
                    {
                        type = parent;
                    }
                    else
                    {
                        alOrderedNodes.Add(elem.SelectSingleNode("mapping:entity/..", nsmgr));
                        ht.Remove(type);
                        type = null;
                    }
                }
            }

            foreach (XmlNode metadata in alOrderedNodes)
                readEntityTyping(domain, metadata, nsmgr, implementationAssembly, generate);
            foreach (XmlNode metadata in alOrderedNodes)
                readPropertiesTyping(domain, metadata, nsmgr, implementationAssembly, generate);
            foreach (XmlNode metadata in alOrderedNodes)
                readReferencesTyping(domain, metadata, nsmgr, implementationAssembly, generate);

            foreach (XmlNode metadata in alOrderedNodes)
                readEntityMapping(domain, metadata, nsmgr, implementationAssembly, generate);
            foreach (XmlNode metadata in alOrderedNodes)
                readPropertiesMapping(domain, metadata, nsmgr, implementationAssembly, generate);
            foreach (XmlNode metadata in alOrderedNodes)
                readReferencesMapping(domain, metadata, nsmgr, implementationAssembly, generate);
        }

      

       
      
       

        [ReflectionPermission(SecurityAction.Demand)]
        private static Type getEntityType(XmlNode metadata, XmlNamespaceManager nsmgr, Assembly assembly)
        {
            XmlNode typeNode = metadata.SelectSingleNode("typing:entity/@type", nsmgr);
            if (typeNode == null)
                throw new Exception("type not found for node " + metadata.Name);
            string type = typeNode.Value.Split(',')[0];
            return assembly.GetType(type, true);
        }

        [ReflectionPermission(SecurityAction.Demand)]
        private static Type getImplementationType(Type entityType, XmlNode metadata, XmlNamespaceManager nsmgr, Assembly implementationAssembly)
        {
            XmlNode node = metadata.ParentNode.SelectSingleNode("typing:entity/@implementationType", nsmgr);
            if (node != null)
                return Type.GetType(node.Value, true);
            return implementationAssembly.GetType(entityType.Namespace + ".Runtime." + entityType.Name + "_", true);
        }

       
        
        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        private static void readEntityTyping(ObjectDomain domain, XmlNode metadata, XmlNamespaceManager nsmgr, Assembly implementationAssembly, bool generate)
        {
            try
            {
                Type entityType;
                if (!generate)
                    entityType = Adoor.Object.Domain.DomainBuild.CodeGeneration.CodeGeneratorCodeDom.getEntityType(metadata, nsmgr);
                else
                    entityType = getEntityType(metadata, nsmgr, implementationAssembly);
                XmlNode isabstract = metadata.SelectSingleNode("typing:entity/@abstract", nsmgr);
                EntityData entityData = new EntityData(domain, entityType, metadata, nsmgr, implementationAssembly, generate);
                //Trace.WriteLine("Registering factory for entity " + entityData.EntityType.Name);
                domain.RegisterEntityData(entityData.EntityType, entityData);
                readKeyFactory(domain, metadata, nsmgr, implementationAssembly, generate);
            }
            catch (TargetInvocationException e)
            {
                throw e.InnerException;
            }
        }

        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        private static void readEntityMapping(ObjectDomain domain, XmlNode metadata, XmlNamespaceManager nsmgr, Assembly implementationAssembly, bool generate)
        {
            XmlNode factoryTypeNode = metadata.SelectSingleNode("mapping:entity/@class", nsmgr);
            if (factoryTypeNode == null)
                throw new Exception("attribute mapping:entity/@class not found for type " + metadata.LocalName);
            try
            {
                Type entityType;
                if (!generate)
                    entityType = Adoor.Object.Domain.DomainBuild.CodeGeneration.CodeGeneratorCodeDom.getEntityType(metadata, nsmgr);
                else
                    entityType = getEntityType(metadata, nsmgr, implementationAssembly);

                XmlNode isabstract = metadata.SelectSingleNode("typing:entity/@abstract", nsmgr);
                EntityData entityData = domain.EntityInfos[entityType];
                Type implementationType = null;

                string[] strs = factoryTypeNode.Value.Split(new char[] { ',' }, 2);
                strs[0] += string.Format("`1[[{0}]]", entityType.AssemblyQualifiedName);
                string factoryTypeName = string.Join(",", strs);
                Type factoryType = Type.GetType(factoryTypeName, true);

                if (generate)
                {
                    implementationType = entityType;
                    if (isabstract != null && isabstract.Value == "true")
                        implementationType = null;
                }
                else
                {
                    if (isabstract == null || isabstract.Value != "true")
                        implementationType = getImplementationType(entityType, metadata, nsmgr, implementationAssembly);
                }
                ConstructorInfo ci = factoryType.GetConstructor(new Type[] { typeof(EntityData), typeof(Type), typeof(XmlNode), typeof(XmlNamespaceManager) });
                if (ci == null)
                    throw new NullReferenceException(" Unable to find a constructor for type " + factoryType.FullName);
                ci.Invoke(new object[] { entityData, implementationType, metadata, nsmgr });
            }
            catch (TargetInvocationException e)
            {
                throw e.InnerException;
            }
        }

        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        private static void readPropertiesTyping(ObjectDomain domain, XmlNode metadata, XmlNamespaceManager nsmgr, Assembly implementationAssembly, bool generate)
        {
            XmlNode typeNode = metadata.SelectSingleNode("typing:entity/@type", nsmgr);
            Type targetType;
            if (!generate)
                targetType = Type.GetType(typeNode.Value, true);
            else
            {
                string type = typeNode.Value.Split(',')[0];
                targetType = implementationAssembly.GetType(type, true);
            }

            EntityData sourceFactory = domain.EntityInfos[targetType];
            XmlNodeList properties = metadata.SelectNodes("*/mapping:property", nsmgr);
            foreach (XmlNode propertyNode in properties)
            {
                // TODO: read propertyName from XML
                string propertyName = propertyNode.ParentNode.LocalName;
                PropertyInfo propertyInfo = sourceFactory.EntityType.GetProperty("m_" + propertyName, BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.FlattenHierarchy | BindingFlags.IgnoreCase);
                if (propertyInfo == null)
                {
                    // second try with the protected case insensitive version
                    propertyInfo = sourceFactory.EntityType.GetProperty(propertyName, BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.FlattenHierarchy | BindingFlags.IgnoreCase);
                }
                if (propertyInfo == null)
                {
                    // third try with the public case sensitive version
                    propertyInfo = sourceFactory.EntityType.GetProperty(propertyName, BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.FlattenHierarchy);
                }

                if (propertyInfo == null)
                    throw new Exception("property " + propertyName + " not found in type " + sourceFactory.EntityType.Name);

                try
                {
                    PropertyData propertyData = new PropertyData(sourceFactory, propertyInfo, propertyNode);
                    //Trace.WriteLine("Registering data for property " + targetType.Name + "." + propertyData.Name);
                }
                catch (TargetInvocationException e)
                {
                    throw e.InnerException;
                }
            }
        }

        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        private static void readPropertiesMapping(ObjectDomain domain, XmlNode metadata, XmlNamespaceManager nsmgr, Assembly implementationAssembly, bool generate)
        {
            XmlNode typeNode = metadata.SelectSingleNode("typing:entity/@type", nsmgr);
            Type targetType;
            if (!generate)
                targetType = Type.GetType(typeNode.Value, true);
            else
            {
                targetType = implementationAssembly.GetType(typeNode.Value.Split(',')[0]);
            }

            EntityData sourceData = domain.EntityInfos[targetType];
            XmlNodeList properties = metadata.SelectNodes("*/mapping:property", nsmgr);
            foreach (XmlNode propertyNode in properties)
            {
                // TODO: read propertyName from XML
                try
                {
                    EntityData entityData = domain.EntityInfos[targetType];
                    PropertyData propertyData = entityData.PropertyInfos[propertyNode.ParentNode.LocalName];
                    XmlNode classTypeNode = propertyNode.SelectSingleNode("@class");
                    string classTypename = classTypeNode != null ? classTypeNode.Value : "Adoor.Object.Relational.PropertyFactory, Adoor.Object.Relational";
                    Type classType = Type.GetType(classTypename, true);
                    ConstructorInfo ci = classType.GetConstructor(new Type[] { typeof(PropertyData), typeof(XmlNode) });
                    if (ci == null)
                        throw new Exception("constructor " + classType.FullName + "::.ctor(PropertyData, XmlNode) not found");
                    IPropertyWriter writer = (IPropertyWriter)ci.Invoke(new object[] { propertyData, propertyNode });
                }
                catch (TargetInvocationException e)
                {
                    throw e.InnerException;
                }
            }
        }

        private static void initProperty(PropertyData propertyData)
        {
            EntityData sourceData = propertyData.SourceData;
            IEntityFactory sourceFactory = (IEntityFactory)sourceData.GetService(typeof(IEntityFactory));
            if (sourceFactory == null)
                return;
            Type implementationType = sourceFactory.ImplementationType;
            string factoryPropertyName = propertyData.Name + "Factory_";
            FieldInfo fi = implementationType.GetField(factoryPropertyName, BindingFlags.Public | BindingFlags.Static | BindingFlags.Instance | BindingFlags.FlattenHierarchy);
            if (fi == null)
                throw new Exception("property " + factoryPropertyName + " not found in type " + implementationType.Name);

            object service = propertyData.GetService(fi.FieldType);
            if (service == null)
                throw new NullReferenceException("Unable to find " + fi.FieldType.FullName + " in " + implementationType.FullName);
            fi.SetValue(null, service);
        }

        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        private static void readReferencesTyping(ObjectDomain domain, XmlNode metadata, XmlNamespaceManager nsmgr, Assembly implementationAssembly, bool generate)
        {
            XmlNode typeNode = metadata.SelectSingleNode("typing:entity/@type", nsmgr);

            Type targetType;
            if (!generate)
                targetType = Type.GetType(typeNode.Value, true);
            else
                targetType = implementationAssembly.GetType(typeNode.Value.Split(',')[0], true);

            EntityData sourceData = domain.EntityInfos[targetType];
            XmlNodeList references = metadata.SelectNodes("*/mapping:reference", nsmgr);
            foreach (XmlNode referenceNode in references)
            {
                string typename = referenceNode.ParentNode.SelectSingleNode("typing:reference/@type", nsmgr).Value;

                Type type;
                if (!generate)
                    type = Type.GetType(typename, true);
                else
                    type = implementationAssembly.GetType(typename.Split(',')[0], true);

                EntityData targetData = domain.EntityInfos[type];
                try
                {
                    XmlNode multiplicityNode = referenceNode.ParentNode.SelectSingleNode("typing:reference/@multiplicity", nsmgr);
                    XmlNode adjmultiplicityNode = referenceNode.ParentNode.SelectSingleNode("typing:reference/@adjacentMultiplicity", nsmgr);
                    string mult = multiplicityNode.Value;
                    string adjmult = adjmultiplicityNode.Value;
                    XmlNode inverseNode = referenceNode.ParentNode.SelectSingleNode("typing:reference/@parent", nsmgr);
                    bool isSingleReference = multiplicityNode == null || multiplicityNode.Value == "1";
                    ReferenceData referenceData = new ReferenceData(referenceNode.ParentNode.LocalName, inverseNode != null ? inverseNode.Value : null, sourceData, targetData, isSingleReference, mult, adjmult);
                    //Trace.WriteLine("Registering factory for reference " + targetType.Name + "." + referenceNode.ParentNode.LocalName);
                    EntityData entityData = domain.EntityInfos[targetType];
                    entityData.RegisterReferenceData(referenceNode.ParentNode.LocalName, referenceData);
                }
                catch (TargetInvocationException e)
                {
                    throw e.InnerException;
                }
            }
        }

        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        private static void readReferencesMapping(ObjectDomain domain, XmlNode metadata, XmlNamespaceManager nsmgr, Assembly implementationAssembly, bool generate)
        {
            XmlNode typeNode = metadata.SelectSingleNode("typing:entity/@type", nsmgr);
            Type sourceType;
            if (!generate)
                sourceType = Type.GetType(typeNode.Value, true);
            else
                sourceType = implementationAssembly.GetType(typeNode.Value.Split(',')[0], true);

            EntityData sourceData = domain.EntityInfos[sourceType];
            XmlNodeList references = metadata.SelectNodes("*/mapping:reference", nsmgr);
            foreach (XmlNode referenceNode in references)
            {
                string typename = referenceNode.ParentNode.SelectSingleNode("typing:reference/@type", nsmgr).Value;
                Type type;
                if (!generate)
                    type = Type.GetType(typename, true);
                else
                    type = implementationAssembly.GetType(typename.Split(',')[0], true);

                string[] strs = referenceNode.SelectSingleNode("@class").Value.Split(new char[] { ',' }, 2);
                strs[0] += string.Format("`1[[{0}]]", type.AssemblyQualifiedName);
                string classTypename = string.Join(",", strs);
                
                Type classType = Type.GetType(classTypename, true);

                ConstructorInfo ci = classType.GetConstructor(new Type[] { typeof(ReferenceData), typeof(XmlNode) });
                if (ci == null)
                    throw new Exception("constructor " + classType.FullName + "::.ctor(ReferenceData, XmlNode) not found");
                EntityData targetFactory = domain.EntityInfos[type];
                try
                {
                    ReferenceData referenceData = sourceData.ReferenceInfos[referenceNode.ParentNode.LocalName];

                    IService service = (IService)ci.Invoke(new object[] { referenceData, referenceNode });
                    //referenceData.RegisterService(typeof(IService), service);
                    service.Initialize();
                }
                catch (TargetInvocationException e)
                {
                    throw e.InnerException;
                }
            }
        }

        private static void initReference(ReferenceData referenceData)
        {
            EntityData sourceData = referenceData.SourceData;
            string propertyName = referenceData.TargetRole + "Factory_";
            IEntityFactory factory = (IEntityFactory)sourceData.GetService(typeof(IEntityFactory));
            if (factory != null && !factory.EntityData.IsAbstract) // otherwise it's an abstract and its descendant will have a factory
            {
                Type implementationType = factory.ImplementationType;
                FieldInfo fi = implementationType.GetField(propertyName, BindingFlags.Public | BindingFlags.Static | BindingFlags.Instance | BindingFlags.FlattenHierarchy);
                if (fi == null)
                    throw new Exception("property " + propertyName + " not found in type " + implementationType.Name);

                object service = referenceData.GetService(fi.FieldType);

                if (service == null)
                    throw new NullReferenceException("Unable to find " + fi.FieldType.FullName + " in " + implementationType.FullName);

                fi.SetValue(null, service);
            }
        }

        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        private static void readKeyFactory(ObjectDomain domain, XmlNode metadata, XmlNamespaceManager nsmgr, Assembly implementationAssembly, bool generate)
        {
            XmlNode typeNode = metadata.SelectSingleNode("typing:entity/@type", nsmgr);
            Type targetType;
            if (!generate)
                targetType = Type.GetType(typeNode.Value, true);
            else
            {
                string sType = typeNode.Value.Split(',')[0];
                targetType = implementationAssembly.GetType(sType, true);
            }

            EntityData sourceFactory = domain.EntityInfos[targetType];
            XmlNode metadata2 = metadata.SelectSingleNode("mapping:entity/mapping:key", nsmgr);
            if (metadata2 == null)
                throw new Exception("tag <mapping:key> not found for " + targetType.Name);
            XmlNode typeNode2 = metadata2.Attributes["class"];
            if (typeNode2 == null)
                throw new Exception("attribute mapping:key/@class not found for " + targetType.Name);
            string keyMappingType = typeNode2.Value;
            if (!keyMappingType.Contains("<"))
                keyMappingType = keyMappingType.Replace(",", string.Format("`1[[{0}]],", targetType.AssemblyQualifiedName));

            Type type = Type.GetType(keyMappingType, true);

            ConstructorInfo ci = type.GetConstructor(new Type[] { typeof(EntityData), typeof(XmlNode) });
            if (ci == null)
                throw new Exception("constructor not found for type " + type.Name);
            try
            {
                sourceFactory.OidData = (IOidData)ci.Invoke(new object[] { sourceFactory, metadata2 });
            }
            catch (TargetInvocationException tie)
            {
                throw tie.InnerException;
            }
            catch (Exception)
            {
                throw;
            }
        }
    }
}
