
using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Security.Permissions;
using System.Xml;
using Adoor.Object.Domain;
using Adoor.Object.Entity;
using Microsoft.CSharp;
using Microsoft.VisualBasic;

namespace Adoor.Object.Domain.DomainBuild.CodeDom
{
    internal class XmlMappingReader : XmlInfoReader
    {
        public static List<XmlNode> SortXmlNode(ObjectDomainXmlMap mapp)
        {
            List<XmlNode> alOrderedNodes = new List<XmlNode>();

            XmlNodeList nodeList = mapp.XmlMap.SelectNodes("//*[count(typing:entity/@parent) = 0]/mapping:entity/..", mapp.NamespaceManager);
            foreach (XmlNode o in nodeList)
                alOrderedNodes.Add(o);

            Dictionary<string, XmlElement> ht = new Dictionary<string, XmlElement>();
            nodeList = mapp.XmlMap.SelectNodes("//*[count(typing:entity/@parent) = 1]", mapp.NamespaceManager);
            foreach (XmlElement elem in nodeList)
            {
                ht[elem.SelectSingleNode("typing:entity/@type", mapp.NamespaceManager).Value] = elem;
            }

            IEnumerator<string> it = null;
            string type = null;
            //TODO: Possible infinite loop ? 
            while (ht.Count > 0)
            {
                if (type == null)
                {
                    it = ht.Keys.GetEnumerator();
                    if (it.MoveNext())
                    {
                        type = (string)it.Current;
                    }
                }
                else
                {
                    XmlElement elem = ht[type];
                    string parent = elem.SelectSingleNode("typing:entity/@parent", mapp.NamespaceManager).Value;
                    if (ht.ContainsKey(parent))
                    {
                        type = parent;
                    }
                    else
                    {
                        alOrderedNodes.Add(elem.SelectSingleNode("mapping:entity/..", mapp.NamespaceManager));
                        ht.Remove(type);
                        type = null;
                    }
                }
            }
            return alOrderedNodes;
        }

        private ObjectDomain _domain = null;
        private ObjectDomainXmlMap _mapp = null;
        private Assembly _implementationAssembly = null;
        private bool _generate = false;

        public XmlMappingReader(ObjectDomain domain, ObjectDomainXmlMap mapp, Assembly implementationAssembly, bool generate)
        {
            _domain = domain;
            _mapp = mapp;
            _implementationAssembly = implementationAssembly;
            _generate = generate;  
        }

        internal void ReadMappingInfo(List<XmlNode> nodes)
        {
            foreach (XmlNode metadata in nodes)
                ReadEntityMapping(metadata);
            foreach (XmlNode metadata in nodes)
                ReadPropertiesMapping(metadata);
            foreach (XmlNode metadata in nodes)
                ReadReferencesMapping(metadata);
        }

        
        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        private void ReadReferencesMapping(XmlNode metadata)
        {
            XmlNamespaceManager nsmgr = _mapp.NamespaceManager;

            XmlNode typeNode = metadata.SelectSingleNode("typing:entity/@type", nsmgr);
            Type sourceType;
            if (!_generate)
                sourceType = Type.GetType(typeNode.Value, true);
            else
                sourceType = _implementationAssembly.GetType(typeNode.Value.Split(',')[0], true);

            EntityData sourceData = _domain.EntityInfos[sourceType];
            XmlNodeList references = metadata.SelectNodes("*/mapping:reference", nsmgr);
            foreach (XmlNode referenceNode in references)
            {
                string typename = referenceNode.ParentNode.SelectSingleNode("typing:reference/@type", nsmgr).Value;
                Type type;
                if (!_generate)
                    type = Type.GetType(typename, true);
                else
                    type = _implementationAssembly.GetType(typename.Split(',')[0], true);

                string[] strs = referenceNode.SelectSingleNode("@class").Value.Split(new char[] { ',' }, 2);
                strs[0] += string.Format("`1[[{0}]]", type.AssemblyQualifiedName);
                string classTypename = string.Join(",", strs);

                Type classType = Type.GetType(classTypename, true);

                ConstructorInfo ci = classType.GetConstructor(new Type[] { typeof(ReferenceData), typeof(XmlNode) });
                if (ci == null)
                    throw new Exception("constructor " + classType.FullName + "::.ctor(ReferenceData, XmlNode) not found");
                EntityData targetFactory = _domain.EntityInfos[type];
                try
                {
                    ReferenceData referenceData = sourceData.ReferenceInfos[referenceNode.ParentNode.LocalName];

                    IService service = (IService)ci.Invoke(new object[] { referenceData, referenceNode });
                    //referenceData.RegisterService(typeof(IService), service);
                    service.Initialize();
                }
                catch (TargetInvocationException e)
                {
                    throw e.InnerException;
                }
            }
        }

        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        private void ReadPropertiesMapping(XmlNode metadata)
        {
            
            XmlNamespaceManager nsmgr = _mapp.NamespaceManager;

            XmlNode typeNode = metadata.SelectSingleNode("typing:entity/@type", nsmgr);
            Type targetType;
            if (!_generate)
                targetType = Type.GetType(typeNode.Value, true);
            else
            {
                targetType = _implementationAssembly.GetType(typeNode.Value.Split(',')[0]);
            }

            EntityData sourceData = _domain.EntityInfos[targetType];
            XmlNodeList properties = metadata.SelectNodes("*/mapping:property", nsmgr);
            foreach (XmlNode propertyNode in properties)
            {
                // TODO: read propertyName from XML
                try
                {
                    EntityData entityData = _domain.EntityInfos[targetType];
                    PropertyData propertyData = entityData.PropertyInfos[propertyNode.ParentNode.LocalName];
                    XmlNode classTypeNode = propertyNode.SelectSingleNode("@class");
                    string classTypename = classTypeNode != null ? classTypeNode.Value : "Adoor.Object.Relational.PropertyFactory, Adoor.Object.Relational";
                    Type classType = Type.GetType(classTypename, true);
                    ConstructorInfo ci = classType.GetConstructor(new Type[] { typeof(PropertyData), typeof(XmlNode) });
                    if (ci == null)
                        throw new Exception("constructor " + classType.FullName + "::.ctor(PropertyData, XmlNode) not found");
                    IPropertyWriter writer = (IPropertyWriter)ci.Invoke(new object[] { propertyData, propertyNode });
                }
                catch (TargetInvocationException e)
                {
                    throw e.InnerException;
                }
            }
        }


        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        private void ReadEntityMapping(XmlNode metadata)
        {
            XmlNamespaceManager nsmgr = _mapp.NamespaceManager;

            XmlNode factoryTypeNode = metadata.SelectSingleNode("mapping:entity/@class", nsmgr);
            if (factoryTypeNode == null)
                throw new Exception("attribute mapping:entity/@class not found for type " + metadata.LocalName);
            try
            {
                Type entityType;
                if (!_generate)
                    entityType = Adoor.Object.Domain.DomainBuild.CodeGeneration.CodeGeneratorCodeDom.getEntityType(metadata, nsmgr);
                else
                    entityType = getEntityType(metadata, nsmgr, _implementationAssembly);

                XmlNode isabstract = metadata.SelectSingleNode("typing:entity/@abstract", nsmgr);
                EntityData entityData = _domain.EntityInfos[entityType];
                Type implementationType = null;

                string[] strs = factoryTypeNode.Value.Split(new char[] { ',' }, 2);
                strs[0] += string.Format("`1[[{0}]]", entityType.AssemblyQualifiedName);
                string factoryTypeName = string.Join(",", strs);
                Type factoryType = Type.GetType(factoryTypeName, true);

                if (_generate)
                {
                    implementationType = entityType;
                    if (isabstract != null && isabstract.Value == "true")
                        implementationType = null;
                }
                else
                {
                    if (isabstract == null || isabstract.Value != "true")
                        implementationType = GetImplementationType(metadata,entityType);
                }
                ConstructorInfo ci = factoryType.GetConstructor(new Type[] { typeof(EntityData), typeof(Type), typeof(XmlNode), typeof(XmlNamespaceManager) });
                if (ci == null)
                    throw new NullReferenceException(" Unable to find a constructor for type " + factoryType.FullName);
                ci.Invoke(new object[] { entityData, implementationType, metadata, nsmgr });
            }
            catch (TargetInvocationException e)
            {
                throw e.InnerException;
            }
        }

        [ReflectionPermission(SecurityAction.Demand)]
        private Type GetImplementationType(XmlNode metadata, Type entityType)
        {
            
            XmlNamespaceManager nsmgr = _mapp.NamespaceManager;

            XmlNode node = metadata.ParentNode.SelectSingleNode("typing:entity/@implementationType", nsmgr);
            if (node != null)
                return Type.GetType(node.Value, true);
            return _implementationAssembly.GetType(entityType.Namespace + ".Runtime." + entityType.Name + "_", true);
        }




    }

}
