
using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Security.Permissions;
using System.Xml;
using Adoor.Object.Domain;
using Adoor.Object.Entity;
using Microsoft.CSharp;
using Microsoft.VisualBasic;

namespace Adoor.Object.Domain.DomainBuild.CodeDom
{
    internal class XmlTypingReader : XmlInfoReader
    {

        private ObjectDomain _domain = null;
        private ObjectDomainXmlMap _mapp = null;
        private Assembly _implementationAssembly = null;
        private bool _generate = false;

        public XmlTypingReader(ObjectDomain domain, ObjectDomainXmlMap mapp, Assembly implementationAssembly, bool generate)
        {
            _domain = domain;
            _mapp = mapp;
            _implementationAssembly = implementationAssembly;
            _generate = generate;  
        }

        internal void ReadTypingInfo(List<XmlNode> nodes)
        {
            foreach (XmlNode metadata in nodes)
                ReadEntityTyping(metadata);
            foreach (XmlNode metadata in nodes)
                ReadPropertiesTyping(metadata);
            foreach (XmlNode metadata in nodes)
                ReadReferencesTyping(metadata);

        }

        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        internal void ReadPropertiesTyping(XmlNode metadata)
        {
            
            XmlNamespaceManager nsmgr = _mapp.NamespaceManager;
            XmlNode typeNode = metadata.SelectSingleNode("typing:entity/@type", nsmgr);
            Type targetType;
            if (!_generate)
                targetType = Type.GetType(typeNode.Value, true);
            else
            {
                string type = typeNode.Value.Split(',')[0];
                targetType = _implementationAssembly.GetType(type, true);
            }

            EntityData sourceFactory = _domain.EntityInfos[targetType];
            XmlNodeList properties = metadata.SelectNodes("*/mapping:property", nsmgr);
            foreach (XmlNode propertyNode in properties)
            {
                // TODO: read propertyName from XML
                string propertyName = propertyNode.ParentNode.LocalName;
                PropertyInfo propertyInfo = sourceFactory.EntityType.GetProperty("m_" + propertyName, BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.FlattenHierarchy | BindingFlags.IgnoreCase);
                if (propertyInfo == null)
                {
                    // second try with the protected case insensitive version
                    propertyInfo = sourceFactory.EntityType.GetProperty(propertyName, BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.FlattenHierarchy | BindingFlags.IgnoreCase);
                }
                if (propertyInfo == null)
                {
                    // third try with the public case sensitive version
                    propertyInfo = sourceFactory.EntityType.GetProperty(propertyName, BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.FlattenHierarchy);
                }

                if (propertyInfo == null)
                    throw new Exception("property " + propertyName + " not found in type " + sourceFactory.EntityType.Name);

                try
                {
                    PropertyData propertyData = new PropertyData(sourceFactory, propertyInfo, propertyNode);
                    //Trace.WriteLine("Registering data for property " + targetType.Name + "." + propertyData.Name);
                }
                catch (TargetInvocationException e)
                {
                    throw e.InnerException;
                }
            }
        }

        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        public void ReadEntityTyping(XmlNode metadata)
        {
            XmlNamespaceManager nsmgr = _mapp.NamespaceManager;

            try
            {
                Type entityType;
                if (!_generate)
                    entityType = Adoor.Object.Domain.DomainBuild.CodeGeneration.CodeGeneratorCodeDom.getEntityType(metadata, nsmgr);
                else
                    entityType = getEntityType(metadata,_mapp.NamespaceManager, _implementationAssembly);
                XmlNode isabstract = metadata.SelectSingleNode("typing:entity/@abstract", nsmgr);
                EntityData entityData = new EntityData(_domain, entityType, metadata, nsmgr, _implementationAssembly, _generate);
                //Trace.WriteLine("Registering factory for entity " + entityData.EntityType.Name);
                _domain.RegisterEntityData(entityData.EntityType, entityData);
                ReadKeyFactory(_domain, metadata,_mapp, _implementationAssembly, _generate);
            }
            catch (TargetInvocationException e)
            {
                throw e.InnerException;
            }
        }

        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        public void ReadReferencesTyping(XmlNode metadata)
        {
            XmlNamespaceManager nsmgr = _mapp.NamespaceManager;

            XmlNode typeNode = metadata.SelectSingleNode("typing:entity/@type", nsmgr);

            Type targetType;
            if (!_generate)
                targetType = Type.GetType(typeNode.Value, true);
            else
                targetType = _implementationAssembly.GetType(typeNode.Value.Split(',')[0], true);

            EntityData sourceData = _domain.EntityInfos[targetType];
            XmlNodeList references = metadata.SelectNodes("*/mapping:reference", nsmgr);
            foreach (XmlNode referenceNode in references)
            {
                string typename = referenceNode.ParentNode.SelectSingleNode("typing:reference/@type", nsmgr).Value;

                Type type;
                if (!_generate)
                    type = Type.GetType(typename, true);
                else
                    type = _implementationAssembly.GetType(typename.Split(',')[0], true);

                EntityData targetData = _domain.EntityInfos[type];
                try
                {
                    XmlNode multiplicityNode = referenceNode.ParentNode.SelectSingleNode("typing:reference/@multiplicity", nsmgr);
                    XmlNode adjmultiplicityNode = referenceNode.ParentNode.SelectSingleNode("typing:reference/@adjacentMultiplicity", nsmgr);
                    string mult = multiplicityNode.Value;
                    string adjmult = adjmultiplicityNode.Value;
                    XmlNode inverseNode = referenceNode.ParentNode.SelectSingleNode("typing:reference/@parent", nsmgr);
                    bool isSingleReference = multiplicityNode == null || multiplicityNode.Value == "1";
                    ReferenceData referenceData = new ReferenceData(referenceNode.ParentNode.LocalName, inverseNode != null ? inverseNode.Value : null, sourceData, targetData, isSingleReference, mult, adjmult);
                    //Trace.WriteLine("Registering factory for reference " + targetType.Name + "." + referenceNode.ParentNode.LocalName);
                    EntityData entityData = _domain.EntityInfos[targetType];
                    entityData.RegisterReferenceData(referenceNode.ParentNode.LocalName, referenceData);
                }
                catch (TargetInvocationException e)
                {
                    throw e.InnerException;
                }
            }
        }


        [ReflectionPermissionAttribute(SecurityAction.Demand)]
        private void ReadKeyFactory(ObjectDomain domain, XmlNode metadata,ObjectDomainXmlMap mapp, Assembly implementationAssembly, bool generate)
        {
            
            XmlNamespaceManager nsmgr = mapp.NamespaceManager;

            XmlNode typeNode = metadata.SelectSingleNode("typing:entity/@type", nsmgr);
            Type targetType;
            if (!generate)
                targetType = Type.GetType(typeNode.Value, true);
            else
            {
                string sType = typeNode.Value.Split(',')[0];
                targetType = implementationAssembly.GetType(sType, true);
            }

            EntityData sourceFactory = domain.EntityInfos[targetType];
            XmlNode metadata2 = metadata.SelectSingleNode("mapping:entity/mapping:key", nsmgr);
            if (metadata2 == null)
                throw new Exception("tag <mapping:key> not found for " + targetType.Name);
            XmlNode typeNode2 = metadata2.Attributes["class"];
            if (typeNode2 == null)
                throw new Exception("attribute mapping:key/@class not found for " + targetType.Name);
            string keyMappingType = typeNode2.Value;
            if (!keyMappingType.Contains("<"))
                keyMappingType = keyMappingType.Replace(",", string.Format("`1[[{0}]],", targetType.AssemblyQualifiedName));

            Type type = Type.GetType(keyMappingType, true);

            ConstructorInfo ci = type.GetConstructor(new Type[] { typeof(EntityData), typeof(XmlNode) });
            if (ci == null)
                throw new Exception("constructor not found for type " + type.Name);
            try
            {
                sourceFactory.OidData = (IOidData)ci.Invoke(new object[] { sourceFactory, metadata2 });
            }
            catch (TargetInvocationException tie)
            {
                throw tie.InnerException;
            }
            catch (Exception)
            {
                throw;
            }

        }

        


    }
}
