#region
/* Adoor.NET - Object/Relational mapping framework leveraging ADO.NET`s powerful disconnected mode.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *  
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *  
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */ 
#endregion

using System.Collections.Generic;
using System.Data;
using System.Data.Common;
using System.Data.Odbc;
using System.Diagnostics;
using Adoor.Relational.SqlDriver;

namespace Adoor.Relational.RelationalSource
{
    [System.ComponentModel.ToolboxItem(true), System.ComponentModel.DesignTimeVisible(true)]    
	public class OdbcRelationalSource: CommonDbRelationalSource
	{
		public OdbcRelationalSource(ISqlDriver sqlDriver, OdbcConnectionStringBuilder connectionStringBuilder):
            base(OdbcFactory.Instance, sqlDriver, connectionStringBuilder)
		{
		}

		protected override void EscapeCommandText(ref string commandText, IDataParameterCollection parameters)
		{
			commandText = commandText.Replace(":?", "?");
			string[] parNames = new string[parameters.Count];
			for(int i = 0; i < parameters.Count; i++)
			{
				parNames[i] = "?";
			}

			commandText = string.Format(commandText, parNames);
		}

		private static bool IsInPrimaryKey(DataColumn dc, DataTable table)
		{
			foreach(DataColumn pk in table.PrimaryKey)
				if (string.Compare(pk.ColumnName, dc.ColumnName, true) == 0)
					return true;
			return false;
		}

		public OdbcCommand GetUpdateCommand(DataTable table, string updateCommand, ICollection<DataColumn> dataColumns)
		{
			OdbcCommand result = (OdbcCommand) dbProviderFactory.CreateCommand();
			foreach (DataColumn dc in dataColumns)
			{
				if (dc.Expression != "" || IsInPrimaryKey(dc, table))
					continue;
				OdbcParameter p = result.Parameters.Add(new OdbcParameter());
				//p.ParameterName = "@p" + (++i).ToString();
				p.SourceColumn = dc.ColumnName;
				p.DbType = Type2DbType(dc.DataType);
			}
			foreach(DataColumn dc in table.PrimaryKey)
			{
				OdbcParameter p = new OdbcParameter();
				//p.ParameterName = "@p" + (++i).ToString();
				p.SourceColumn = dc.ColumnName;
				p.SourceVersion = DataRowVersion.Original;
				p.DbType = Type2DbType(dc.DataType);
				result.Parameters.Add(p);
			}
			if (updateWhereAll)
				foreach (DataColumn dc in dataColumns)
				{
					OdbcParameter p = new OdbcParameter();
					//p.ParameterName = "@p" + (++i).ToString();
					p.SourceColumn = dc.ColumnName;
					p.SourceVersion = DataRowVersion.Original;
					p.DbType = Type2DbType(dc.DataType);
					result.Parameters.Add(p);
				}
			EscapeCommandText(ref updateCommand, result.Parameters);
			result.CommandText = updateCommand;
			return result;
		}

		public OdbcCommand GetInsertCommand(DataTable table, string insertCommand, ICollection<DataColumn> dataColumns)
		{
			OdbcCommand result = (OdbcCommand) dbProviderFactory.CreateCommand();
			foreach (DataColumn dc in dataColumns)
			{
				if (dc.Expression != "")
					continue;
				OdbcParameter p = new OdbcParameter();
				//p.ParameterName = "@p" + (++i).ToString();
				p.SourceColumn = dc.ColumnName;
				p.DbType = Type2DbType(dc.DataType);
				result.Parameters.Add(p);
			}
			EscapeCommandText(ref insertCommand, result.Parameters);
			result.CommandText = insertCommand;
			return result;
		}

		public OdbcCommand GetDeleteCommand(DataTable table, string deleteCommand, ICollection<DataColumn> dataColumns)
		{
			OdbcCommand result = (OdbcCommand) dbProviderFactory.CreateCommand();
			foreach(DataColumn dc in table.PrimaryKey)
			{
				OdbcParameter p = new OdbcParameter();
				//p.ParameterName = "@p" + (++i).ToString();
				p.SourceColumn = dc.ColumnName;
				p.SourceVersion = DataRowVersion.Original;
				p.DbType = Type2DbType(dc.DataType);
				result.Parameters.Add(p);
			}
			if (updateWhereAll) 
			{
				foreach (DataColumn dc in dataColumns)
				{
					OdbcParameter p = new OdbcParameter();
					//p.ParameterName = "@p" + (++i).ToString();
					p.SourceColumn = dc.ColumnName;
					p.SourceVersion = DataRowVersion.Original;  // ignored ?
					p.DbType = Type2DbType(dc.DataType);
					result.Parameters.Add(p);
				}
			}
			EscapeCommandText(ref deleteCommand, result.Parameters);
			result.CommandText = deleteCommand;
			return result;
		}

		public override DbDataAdapter GetDataAdapter(DataTable table,
			string selectCommand, string updateCommand, string insertCommand, string deleteCommand,
			ICollection<DataColumn> dataColumns)
		{

            OdbcDataAdapter result = (OdbcDataAdapter) dbProviderFactory.CreateDataAdapter();
            result.SelectCommand = (OdbcCommand) dbProviderFactory.CreateCommand();
            result.SelectCommand.CommandText = selectCommand;
            result.UpdateCommand.Connection = (OdbcConnection)dbProviderFactory.CreateConnection();
            result.UpdateCommand.Connection.ConnectionString = ConnectionStringBuilder.ConnectionString;
            if (AdoorLoger.IsDebugEnabled)
                result.RowUpdating += new OdbcRowUpdatingEventHandler(OnRowUpdating);

			// Update command
			result.UpdateCommand = GetUpdateCommand(table, updateCommand, dataColumns);
            result.UpdateCommand.Connection = result.UpdateCommand.Connection;
				
			// Insert command
			result.InsertCommand = GetInsertCommand(table, insertCommand, dataColumns);
            result.InsertCommand.Connection = result.UpdateCommand.Connection;
			
			// Delete command
			result.DeleteCommand = GetDeleteCommand(table, deleteCommand, dataColumns);
            result.DeleteCommand.Connection = result.UpdateCommand.Connection;
			return result;
		}

		public void OnRowUpdating(object sender, OdbcRowUpdatingEventArgs args)
		{
            AdoorLoger.Debug(Tracer.GetTrace(args.Command.CommandText, args.Command));			
		}
	}
}
