#region
/* Adoor.NET - Object/Relational mapping framework leveraging ADO.NET`s powerful disconnected mode.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *  
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *  
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */ 
#endregion

using System;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Xsl;
using System.IO;

namespace Adoor.Metadata
{
	/// <summary>
	/// Summary description for Class1.
	/// </summary>
	public class MetadataManager
	{
		protected XmlDocument document;

		public MetadataManager(String url)
		{
			this.document = LoadDocument(url);
			
			ConsolidateInherit();
		}

		public XmlDocument Document
		{
			get { return this.document; }
		}

		protected XmlDocument LoadDocument(String url)
		{
			XmlDocument result = new XmlDocument();
			result.Load(url);
			XmlNamespaceManager nsmanager = new XmlNamespaceManager(new NameTable());
			nsmanager.AddNamespace("meta", "urn:xmd-meta");
			XmlNodeList nodeList = result.SelectNodes("//meta:import", nsmanager);
			foreach(XmlNode node in nodeList)
			{
				XmlNode hrefNode = node.SelectSingleNode("@href");
				String href = hrefNode.Value;
				if (!href.StartsWith("/") && !href.StartsWith("\\") & href.IndexOf(':') == -1) 
				{
					int i = url.LastIndexOf('/');
					int j = url.LastIndexOf('\\');
					if (j > i) i = j;
					if (i != -1)
						href= url.Substring(0, i + 1) + href;
				}
				XmlDocument import = LoadDocument(href);
				XmlNode parentNode = node.ParentNode;
				parentNode.RemoveChild(node);
				ConsolidateImport(parentNode, import.DocumentElement);
			}
			return result;
		}

		protected void ConsolidateImport(XmlNode parentNode, XmlNode importNode)
		{
			foreach(XmlNode childNode in parentNode.ChildNodes) 
			{
				if (childNode.LocalName == importNode.LocalName && childNode.NamespaceURI == importNode.NamespaceURI) 
				{
					// TODO: consolidate attributes ?
					foreach(XmlNode importChildNode in importNode.ChildNodes)
						ConsolidateImport(childNode, importChildNode);
					return;
				}
			}
			parentNode.AppendChild(parentNode.OwnerDocument.ImportNode(importNode, true));
		}

		protected void ConsolidateInherit()
		{
			XmlNamespaceManager nsmanager = new XmlNamespaceManager(new NameTable());
			nsmanager.AddNamespace("meta", "urn:xmd-meta");
			XmlNodeList nodeList = this.document.SelectNodes("//meta:inherit", nsmanager);
			do 
			{
				foreach(XmlNode node in nodeList)
				{
					XmlNode selectNode = node.SelectSingleNode("@select");
					String path = selectNode.Value;
					
					XmlNode filterNode = node.SelectSingleNode("@filter");
					String filter = filterNode == null ? "*" : filterNode.Value;


					XmlNamespaceManager nsmanager2 = new XmlNamespaceManager(new NameTable());
					String metaPrefix = node.GetPrefixOfNamespace("urn:xmd-meta");
					nsmanager2.AddNamespace(metaPrefix, "urn:xmd-meta");
					XmlNode rootNode = path.StartsWith("/") ? this.document : node.ParentNode;
					XmlNodeList inheritList = rootNode.SelectNodes(path + "/" + filter, GetNamespaceManager(node));

					XmlNode parentNode = node.ParentNode;
					parentNode.RemoveChild(node);
					
					foreach(XmlNode inheritNode in inheritList)
					{
						parentNode.AppendChild(inheritNode.Clone());
					}
				}
				nodeList = this.document.SelectNodes("//meta:inherit", nsmanager);
			} while (nodeList.Count != 0);
		}

		public virtual string GetPrefixOfNamespace(XmlNode node, string namespaceURI) 
		{
			XmlDocument document = node.NodeType == XmlNodeType.Document ? (XmlDocument)node : node.OwnerDocument;
			if (namespaceURI == "http://www.w3.org/2000/xmlns/")
				return "xmlns";
			if (document != null) 
			{
				namespaceURI = document.NameTable.Add(namespaceURI);
				while (node != null) 
				{
					switch (node.NodeType) 
					{
						case  XmlNodeType.Element:
							if (Equal(node.NamespaceURI, namespaceURI))
								return node.Prefix;
							XmlElement element = (XmlElement) node;
							if (element.HasAttributes) 
							{
								for(int i = element.Attributes.Count - 1; i >= 0; --i)
								{
									XmlNode attr = element.Attributes[i];
									if (Equal(attr.Prefix, "xmlns") && Equal(attr.Value, namespaceURI))
										return attr.LocalName;
								}
							}
							node = node.ParentNode;
							break;
						case XmlNodeType.Attribute:
							node = ((XmlAttribute) node).OwnerElement;
							break;
						default:
							node = node.ParentNode;
							break;
					}
				}
			}
			return String.Empty;
		}


		public virtual string GetNamespaceOfPrefix(XmlNode node, string prefix) 
		{
			XmlDocument document = node.NodeType == XmlNodeType.Document ? (XmlDocument)node : node.OwnerDocument;
			if (document != null) 
			{
				prefix = document.NameTable.Get(prefix);
				if (prefix == null)
					return String.Empty;
				while (node != null) 
				{
					switch (node.NodeType) 
					{
						case XmlNodeType.Element:
							if (Equal(node.Prefix, prefix))
								return node.NamespaceURI;
							XmlElement element = (XmlElement) node;
							if (element.HasAttributes) 
							{
								for(int i = element.Attributes.Count - 1; i >= 0; --i)
								{
									XmlNode attr = element.Attributes[i];
									if (Equal(attr.Prefix, "xmlns") && Equal(attr.LocalName, prefix))
										return attr.Value;
								}
							}
							node = node.ParentNode;
							break;
						case XmlNodeType.Attribute:
							node = ((XmlAttribute) node).OwnerElement;
							break;
						default:
							node = node.ParentNode;
							break;
					}
				}
			}
			return String.Empty;
		}

		public static XmlNamespaceManager GetNamespaceManager(XmlNode node) 
		{
			XmlNamespaceManager result = new XmlNamespaceManager(new NameTable());
			//result.AddNamespace("xmlns", "http://www.w3.org/2000/xmlns/");
			while (node != null) 
			{
				switch (node.NodeType) 
				{
					case XmlNodeType.Element:
						XmlElement element = (XmlElement) node;
						if (element.HasAttributes) 
						{
							for(int i = element.Attributes.Count - 1; i >= 0; --i)
							{
								XmlNode attr = element.Attributes[i];
								if (Equal(attr.Prefix, "xmlns"))
									if (!result.HasNamespace(attr.LocalName))
										result.AddNamespace(attr.LocalName, attr.Value);
							}
						}
						node = node.ParentNode;
						break;
					case XmlNodeType.Attribute:
						node = ((XmlAttribute) node).OwnerElement;
						break;
					default:
						node = node.ParentNode;
						break;
				}
			}
			return result;
		}


		public static bool Equal(string strA, string strB) 
		{
			return (object)strA == (object)strB;
		}

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			MetadataManager manager = new MetadataManager(@"..\..\biblio.xmd");

			XmlTextWriter writer = new XmlTextWriter(Console.Out);
			writer.Formatting = Formatting.Indented;
			manager.document.WriteTo(writer);
			writer.Flush();
			Console.WriteLine();
		}

	}
}
