using System;
using System.Collections;
using System.Text;
using System.IO;
using System.Diagnostics;
using System.Data;
using Adoor.Object;
using Adoor.Object.Query;
using NUnit.Framework;

using Adoor.Ouvrages;
using Adoor.Emprunts;
using Adoor.Object.Disconnected;

namespace Biblio.NUnit.Inheritance
{
	[TestFixture]
	public class TestLazy : BaseTest
	{
		public ObjectContext Context;

		[SetUp]
		public void Setup()
		{
			Context = CreateContext();

			// Redirect the trace to the NUnit window
			Trace.Listeners.Add(new TextWriterTraceListener(Console.Out, "Console"));
		}

		[TearDown]
		public void TearDown()
		{
			Trace.Listeners.Remove("Console");
		}

		[Test]
		public void TestForeignKeyFactory()
		{
			Exemplaire exemplaire = CreateExemplaire(Context);
			string exemplaireOid = exemplaire.ToString();
			Livre livre = CreateLivre(Context);
			exemplaire.Ouvrage = livre;
			string livreOid = livre.ToString();
			Context.Update();
			Context.LocalStorage.Clear();
			Exemplaire exemplaireTest = (Exemplaire)Context.Find(typeof(Exemplaire), exemplaireOid);
			Assert(exemplaireTest != null);
			Ouvrage ouvrageTest = (Ouvrage)Context.Find(typeof(Ouvrage), livreOid);
			Assert(ouvrageTest != null);
			Assert(exemplaireTest.Ouvrage == ouvrageTest);
		}

		[Test]
		public void CollectAfterCreate()
		{
			Context.RemoteSource.DeleteAll(typeof(Ouvrage));
			Ouvrage ouvrage1 = CreateLivre(Context);
			string ouvrage1oid = ouvrage1.ToString();
			Ouvrage ouvrage2 = CreateLivre(Context);
			string ouvrage2oid = ouvrage2.ToString();
			Ouvrage ouvrage3 = CreateNumeroRevue(Context);
			string ouvrage3oid = ouvrage3.ToString();
			Context.Update();
			Context.LocalStorage.Clear();
			Context.Preload(typeof(Ouvrage));
			ICollection ouvrages = Context.LocalStorage.Collect(typeof(Ouvrage));
			Assert(ouvrages.Count == 3);
			int ouvrage1count = 0;
			int ouvrage2count= 0;
			foreach(Ouvrage o in ouvrages)
			{
				Assert(o.ToString() == ouvrage1oid || o.ToString() == ouvrage2oid || o.ToString() == ouvrage3oid);
				if (o.ToString() == ouvrage1oid)
					++ouvrage1count;
				if (o.ToString() == ouvrage2oid)
					++ouvrage2count;
			}
			Assert(ouvrage1count == 1);
			Assert(ouvrage2count == 1);
		}

		[Test]
		public void Delete()
		{
			Ouvrage ouvrage1 = CreateLivre(Context);
			Ouvrage ouvrage2 = CreateLivre(Context);
			Ouvrage ouvrage3 = CreateNumeroRevue(Context);
			Context.Update();
			Context.LocalStorage.Clear();
			Context.RemoteSource.DeleteAll(typeof(Ouvrage));
			Assert(Context.RemoteSource.GetCount(typeof(Ouvrage)) == 0);
		}

		[Test]
		public void ComplexQuery1()
		{
			Ouvrage ouvrage1 = CreateLivre(Context);
			Ouvrage ouvrage2 = CreateLivre(Context);
			Ouvrage ouvrage3 = CreateNumeroRevue(Context);
			Bibliotheque b = CreateBibliotheque(Context);

			// premier exemplaire de ouvrage1
			Exemplaire exemplaire1 = CreateExemplaire(Context);
			exemplaire1.Ouvrage = ouvrage1;
			b.ListeExemplairesAchetes.Add(exemplaire1);

			// deuxime exemplaire de ouvrage1
			Exemplaire exemplaire2 = CreateExemplaire(Context);
			exemplaire2.Ouvrage = ouvrage1;
			b.ListeExemplairesAchetes.Add(exemplaire2);

			// premier exemplaire de ouvrage3
			Exemplaire exemplaire3 = CreateExemplaire(Context);
			exemplaire3.Ouvrage = ouvrage3;
			b.ListeExemplairesAchetes.Add(exemplaire3);

			Context.Update();

			// 2 ouvrages disponibles dans la bibliothque (ouvrage1 et ouvrage3);

			// Ouvrage means "the key columns of Ouvrage"
			Assert(2 == Context.RemoteSource.GetCount("select distinct Ouvrage from Bibliotheque b, Exemplaire in b.ListeExemplairesAchetes, Exemplaire.Ouvrage where b = ?", b));
		}

		[Test, ExpectedException(typeof(Exception))]
		public void TestErreur1()
		{
			Ouvrage ouvrage1 = CreateLivre(Context);
			Bibliotheque b = CreateBibliotheque(Context);
			b.ListeExemplairesAchetes.Add(ouvrage1); // erreur : Add(exemplaire1)
		}

		[Test]
		public void Preload()
		{
			Context.RemoteSource.DeleteAll(typeof(Ouvrage));
			Context.RemoteSource.DeleteAll(typeof(Editeur));
			Ouvrage ouvrage1 = CreateLivre(Context);
			string ouvrage1oid = ouvrage1.ToString();
			Ouvrage ouvrage2 = CreateLivre(Context);
			string ouvrage2oid = ouvrage2.ToString();
			Ouvrage ouvrage3 = CreateNumeroRevue(Context);
			string ouvrage3oid = ouvrage3.ToString();
			Editeur editeur1 = CreateEditeur(Context);
			Editeur editeur2 = CreateEditeur(Context);
			ouvrage1.Editeur = editeur1;
			ouvrage2.Editeur = editeur2;
			ouvrage3.Editeur = editeur1;
			Context.Update();
			Context.LocalStorage.Clear();
			Context.Preload("Ouvrage, Ouvrage.Editeur");
		}

	}
}
